// WGet.java
// $Id: JWGet.java,v 1.10 1999/11/09 10:17:53 bmahe Exp $
// (c) COPYRIGHT MIT, INRIA and Keio, 1999.
// Please first read the full copyright statement in file COPYRIGHT.html
package org.w3c.jwput;
 
import java.io.*;
import java.net.*;
import java.util.*;

import org.w3c.tools.codec.*;
import org.w3c.www.protocol.http.*;
import org.w3c.www.http.*;

/**
 * Get files on the Web
 * @version $Revision: 1.10 $
 * @author  Benot Mah (bmahe@w3.org)
 */
public class JWGet extends JW {

    int tries    = 0;

    /**
     * Get the backup file relative to the given file.
     * @param the file to backup
     * @return the backup file
     */
    protected File getBackupFile(File file) {
        return new File(file.getParent(), file.getName()+".bak");
    }

    /**
     * Get the latest version of the remote file. (and retry if connection
     * is closed). display the banner.
     * @param toget the file to update
     * @param furl the URL
     * @param backup if true the local file will be backuped
     */
    public void get(File toget, String furl, boolean backup) 
	throws IOException
    {
	if (! quiet) {
	    ui.displayMessage(BANNER);
	    ui.displayMessage("Version: "+VERSION);
	}
	getFile(toget, furl, backup);
    }

    /**
     * Get the latest version of the remote file. (and retry if connection
     * is closed).
     * @param toget the file to update
     * @param furl the URL
     * @param backup if true the local file will be backuped
     */
    public File getFile(File toget, String furl, boolean backup) 
	throws IOException
    {
	return getFile(toget, furl, backup, null);
    }

    /**
     * Get the latest version of the remote file. (and retry if connection
     * is closed).
     * @param toget the file to update
     * @param furl the URL
     * @param backup if true the local file will be backuped
     */
    public File getFile(File toget, 
			String furl, 
			boolean backup, 
			Hashtable headers) 
	throws IOException
    {
	HttpRange range    = null;
	int       numtries = tries;
	do {
	    try {
		return getFile(toget, furl, backup, true, range, headers);
	    } catch (InterruptedGetException ex) {
		ui.errlog("Connection closed at byte "+
			  ex.bytesTransferred+". Retrying ...");
		ui.sync();
 		range =  HttpFactory.makeRange("bytes", 
					       (int)ex.bytesTransferred,
					       (int)ex.bytesExpected-1);
		backup = false;
	    }
	} while (--numtries > -1);
	return null;
    }

    /**
     * Get the latest version of the remote file. (no retry)
     * @param toget the file to update
     * @param furl the URL
     * @param backup if true the local file will be backuped
     * @param trace should we display trace?
     * @param range the HttpRange to request
     */
    public File getFile(File toget, 
			String furl,
			boolean backup,
			boolean trace,
			HttpRange range,
			Hashtable headers) 
	throws IOException
    {
	if (! quiet && trace) {
	    ui.log("Downloading "+furl+" ...");
	    ui.sync();
	}
	//headers...
	if (headers == null)
	    headers = new Hashtable(10);
	if (range != null)
	    headers.put("Range",range.toString());
	//get the file
	Reply reply = performRequest("GET", furl, headers);
	if (reply == null)
	    return null;
	//reply
	if (reply.getStatus()/100 == 2) {
	    if ((toget == null) || toget.isDirectory()) {
		String file = getFileName(furl);
		if (file == null) {
		    String location = reply.getContentLocation();
		    if (location != null)
			file = getFileName(location);
		    if (file == null)
			file = getIdxFileName(furl);
		}
		if (toget == null)
		    toget = new File(file);
		else 
		    toget = new File(toget, file);
	    }
	    //backup the file
	    if ((backup) && (toget.exists()))
		org.w3c.util.IO.copy(toget, getBackupFile(toget));
	    //get the remote file
	    InputStream in = reply.getInputStream();

	    if (! quiet)
		in = ui.getMonitoredInputStream(in, "Downloading "+
						toget.getName(), 
						reply.getContentLength());
	    FileOutputStream out = null;
	    try {
		//append?
		out =  new FileOutputStream(toget.getCanonicalPath(), 
					    (range != null));
		org.w3c.util.IO.copy(in, out);
	    } catch (IOException ex) {
		out.close();
		//connection closed
		InterruptedGetException iex = 
		    new InterruptedGetException(ex.getMessage());
		iex.bytesTransferred = toget.length();
		if (range != null) {
		    iex.bytesExpected = range.getLastPosition()+1;
		} else {
		    iex.bytesExpected = reply.getContentLength();
		}
		throw iex;
	    }
	    //save etag
	    HttpEntityTag remoteETag = reply.getETag();
	    if (remoteETag != null) {
		String etag = remoteETag.getTag();
		saveEtag(toget, furl, remoteETag.getTag());
	    }
	    if (! quiet) {
		String array[] = { toget.getName()+" downloaded at",
				   toget.getCanonicalPath() };
		ui.displayMessage(array, "File downloaded");
	    }
	    return toget;
	} else if (reply.getStatus() == HTTP.NOT_MODIFIED) {
	    skipBody(reply);
	    ui.displayMessage("You already have the latest version of "+
			      toget.getName());
	    return toget;
	} else {
	    displayErrorMessage(reply);
	    return null;
	}
    }

    /**
     * Update the user properties
     * @param props the new peoperties
     */
    public void updateProperties(Properties props) {
	super.updateProperties(props);
	this.manager = getManager(props);
    }

    protected void initialize(Properties props) {
	super.initialize(props);

	String numtries = props.getProperty(TRIES_P);
	try {
	    this.tries = 
		(numtries == null) ? DEF_TRIES : Integer.parseInt(numtries);
	} catch (Exception ex) {
	    this.tries = DEF_TRIES;
	}
    }

    /**
     * Constructor
     * @param manager the HttpManager to use
     */
    public JWGet(HttpManager manager) {
	this(manager, new org.w3c.jwput.tui.JWTUI());
    }

    /**
     * Constructor
     * @param manager the HttpManager to use
     * @param ui the User Interface to use
     * @see org.w3c.jwput.gui.JWGUI
     * @see org.w3c.jwput.tui.JWTUI
     */
    public JWGet(HttpManager manager, JWUI ui) {
	this.ui      = ui;
	this.manager = manager;
	initialize(manager.getProperties());
    }

    /**
     * Constructor
     * @param userprops the Properties to use
     */
    public JWGet(Properties userprops) {
	this(new org.w3c.jwput.tui.JWTUI(), userprops);
    }

    /**
     * Constructor
     * @param ui the User Interface to use
     * @see org.w3c.jwput.gui.JWGUI
     * @see org.w3c.jwput.tui.JWTUI
     * @param userprops the Properties to use
     */
    public JWGet(JWUI ui, Properties userprops) {
	this.ui = ui;
	Properties props = new Properties();
	File file = getPreferencesFile();
	try {
	    InputStream in = 
		new BufferedInputStream(new FileInputStream(file));
	    try {
		props.load(in);
	    } catch (IOException ex) {
		ui.displayError("unable to load "+WPUTRC);
		ui.errlog(ex.getMessage());
		ui.sync();
	    }
	} catch (FileNotFoundException ex) {
	    ui.errlog(file.getAbsolutePath()+" not found");
	}
	//merge user properties with default properties
	Enumeration enum = userprops.propertyNames();
	while (enum.hasMoreElements()) {
	    String key = (String)enum.nextElement();
	    props.put(key, userprops.getProperty(key));
	}
	initialize(props);
	this.manager = getManager(props);
    }
    
    /**
     * Display the "Usage" message.
     */
    public static void usage() {
	System.out.println("JWGet "+VERSION+", a HTTP/1.1 GET tool.");
	System.out.println("Usage: java JWGet [OPTION] <url>");
	System.out.println("\nOptions:\n");
	System.out.println("  -h                     display this message.");
	System.out.println("  -header_XXX <value>    specify a HTTP header");
	System.out.println("  -no-backup             disable backuping.");
	System.out.println("  -no-proxy              direct connection.");
	System.out.println("  -o <file>              specify the output file");
	System.out.println("  -prompt                prompt monde on.");
	System.out.println("  -user <username>       Auth: the user name.");
	System.out.println("  -password <password>   Auth: the password.");
	System.out.println("  -proxy-host <host>     eg: proxy.foobar.com");
	System.out.println("  -proxy-port <port>     eg: 81 (default value)");
	System.out.println("  -quiet                 quiet mode on.");
	System.out.println("  -tries <number>        set number of retries "+
			   "(default is 3).");
	System.out.println("  -verbose               verbose mode on.");
	System.out.println("\nMail bug reports to <bmahe@w3.org>.");
	System.exit(1);
    }

    /**
     * Display the message for bad command line error.
     * @param msg a message.
     */
    public static void badArgs(String msg) {
	System.out.println("JWGet: "+msg);
	System.out.println("Usage: java JWGet [OPTION] <url>");
	System.out.println("\n'JWGet -h' for more options.");
	System.exit(1);
    }

    public static String getIdxFileName(String url) {
	int     len      = url.length();
	int     endidx   = len;
	int     startidx = 0;
	boolean dir      = false;

	if (url.endsWith("/")) {
	    endidx--;
	    dir = true;
	}
	startidx = url.lastIndexOf('/', endidx-1);
	String name = null;
	if (startidx != -1)
	    name = url.substring(startidx+1, endidx);
	if (dir)
	    name += ".idx";
	return name;
    }

    public static String getFileName(String url) {
	int     len      = url.length();
	int     endidx   = len;
	int     startidx = 0;

	if (url.endsWith("/"))
	    return null;

	startidx = url.lastIndexOf('/', endidx-1);
	String name = null;
	if (startidx != -1)
	    name = url.substring(startidx+1, endidx);
	return name;
    }

    public static void main(String args[]) {
	String  output  = null;
	String  url     = null;
	boolean backup = true;

	Properties props = new Properties();
	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-user")) {
		props.put(USER_P, args[++i]);
	    } else if (args[i].equals("-password")) {
		props.put(PASSWORD_P, args[++i]);
	    } else if (args[i].equals("-no-proxy")) {
		props.put(PROXY_SET_P, FALSE_P);
	    } else if (args[i].equals("-proxy-host")) {
		props.put(PROXY_HOST_P, args[++i]);
	    } else if (args[i].equals("-proxy-port")) {
		props.put(PROXY_PORT_P, args[++i]);
	    } else if (args[i].equals("-verbose")) {
		props.put(VERBOSE_P, TRUE_P);
	    } else if (args[i].equals("-quiet")) {
		props.put(QUIET_P, TRUE_P);
	    } else if (args[i].equals("-prompt")) {
		props.put(PROMPT_P, TRUE_P);
	    } else if (args[i].equals("-no-backup")) {
		backup = false;
	    } else if (args[i].equals("-tries")) {
		props.put(TRIES_P, args[++i]);
	    } else if (args[i].equals("-h")) {
		usage();
	    } else if (args[i].equals("-o")) {
		output = args[++i];
	    } else if (args[i].startsWith("-"+HEADER_PREFIX)) {
		props.put(args[i].substring(1), args[++i]);
	    } else {
		if (url == null)
		    url = args[i];
		else
		    usage();
	    }
	}
	if (url == null)
	    badArgs("Missing file");
	//ok let's go...
	JWGet wget = new JWGet(props);
	File  file = null;
	if (output != null)
	    file = new File(output);
	try {
	    wget.get(file, url, backup);
	} catch (IOException ex) {
	    System.err.println("\nError : "+ex.getMessage());
	}
    }
}
