// JWPut.java
// $Id: JWPut.java,v 1.17 1999/11/09 10:17:53 bmahe Exp $
// (c) COPYRIGHT MIT, INRIA and Keio, 1999.
// Please first read the full copyright statement in file COPYRIGHT.html
package org.w3c.jwput;
 
import java.io.*;
import java.net.*;
import java.util.*;

import org.w3c.tools.codec.*;
import org.w3c.www.protocol.http.*;
import org.w3c.www.mime.*;
import org.w3c.www.http.*;

/**
 * Put files on the web
 * @version $Revision: 1.17 $
 * @author  Benot Mah (bmahe@w3.org)
 */
public class JWPut extends JW {
    
    String      mime      = null;
    String      mimedef   = null;
    boolean     overwrite = false;
    boolean     rec       = false;
    JWGet       wget      = null;

    /**
     * Create a HTTP/1.1 request
     * @return a Request instance
     */
    public Request createRequest() {
	Request request = manager.createRequest();
	request.setExpect("100-continue");
	request.setObserver(this);
	return request;
    }

    /**
     * Guess a mime type for the given file.
     * @param name the file name
     * @return a String instance
     */ 
    public String getMimeType(String name) {
	return getMimeType(name, prompt);
    }

    /**
     * Guess a mime type for the given file.
     * @param name the file name
     * @param prompt should we prompt the user for a mime type if we can't 
     * guess it?
     * @return a String instance
     */ 
    public String getMimeType(String name, boolean prompt) {
	Properties props   = getProperties();
	int        idx     = name.lastIndexOf('.');
	String     ext     = "noext";
	String     mime    = null;
	String     mimedef = props.getProperty(MIMEDEF_P);

	if (idx != -1)
	    ext = name.substring(idx+1);
	//user mime types
	mime = props.getProperty(ext);
	//predifined mime types
	if (mime == null) {
	    mime = Utils.guessContentTypeFromName(name);
	    //still nothing? try default mime type
	    if ((mime == null) || (mime.equals("content/unknown"))) {
		if (mimedef != null) {
		    ui.errlog("Unknown mime type for "+name);
		    ui.errlog("Using default: "+mimedef+"\n");
		    ui.sync();
		    mime = mimedef;
		} else if (prompt) {
		    ui.sync();
		    //nothing? sure? ok prompt
		    String message[] = {
			"Unknown mime-type for "+name,
			"Please enter mime type: "
		    };
		    mime = ui.getString(message);
		} else {
		    mime = null;
		}
	    }
	}
	return mime;
    }

    /**
     * Perform a HTTP/1.1 PUT
     * @param toput The file to upload
     * @param puturl the location url
     * @param user auth: username
     * @param password auth: password
     * @return a Reply instance
     * @exception FileNotFoundException if toput  is not found
     */
    public Reply put(File toput, String puturl, String etag) 
	throws FileNotFoundException
    {
	return put(toput, puturl, etag, null);
    }
    /**
     * Perform a HTTP/1.1 PUT
     * @param toput The file to upload
     * @param puturl the location url
     * @param user auth: username
     * @param password auth: password
     * @return a Reply instance
     * @exception FileNotFoundException if toput  is not found
     */
    public Reply put(File toput, 
		     String puturl, 
		     String etag, 
		     Hashtable headers) 
	throws FileNotFoundException
    {
	if (headers == null)
	    headers = new Hashtable(10);
	//
	// mime type
	//
	if (headers.get("Content-Type") == null) {
	    String mymime = null;
	    if (mime == null)
		mymime = getMimeType(toput.getName());
	    else
		mymime = mime;
	    if (mymime == null)
		throw new RuntimeException("No mime type, aborted.");
	    try {
		//check mime type
		new MimeType(mymime);
	    } catch (MimeTypeFormatException ex) {
		throw new RuntimeException("Invalid mime type : "+mymime);
	    }
	    headers.put("Content-Type", mymime);
	}
	//
	// If match
	//
	if (etag != null)
	    headers.put("If-Match", "\""+etag+"\"");
	//
	// body
	//
	InputStream in = new BufferedInputStream(new FileInputStream(toput));
	long len = toput.length();
	if (! quiet)
	    in = ui.getMonitoredInputStream(in, "Uploading "+toput.getName(),
					    (int)len);
	headers.put("Content-Length", String.valueOf(len));
	headers.put("Expect", "100-Continue");
	//run the request
	return performRequest("PUT", puturl, in, headers);
    }

    /**
     * Perform the put(s)
     * @exception FileNotFoundException if a file is not found
     */
    public void perform(String file, String url)
	throws FileNotFoundException
    {
	if (! quiet) {
	    ui.displayMessage(BANNER);
	    ui.displayMessage("Version: "+VERSION);
	}
	File toput = new File(file);
	if (toput.isDirectory() && rec) {
	    if (! url.endsWith("/"))
		url += "/";
	    putDir(toput, url);
	} else
	    putFile(toput, url);
    }

    /**
     * Perform a DELETE
     */
    public void delete(String url) {
	if (rec)
	    throw new RuntimeException("Recurvise delete not allowed");
	Reply reply = performRequest("DELETE", url);
	if (reply == null)
	    return;
	skipBody(reply);
	if (! quiet) {
	    if (reply.getStatus() / 100 == 2) {
		ui.displayMessage(url+" deleted.");
	    } else {
		//error
		displayErrorMessage(reply);
	    }
	}
    }

    /**
     * Put (recursively) the files to the location URL.
     * @param dir the directory
     * @param durl the location URL
     */
    public void putDir(File dir, String durl) 
	throws FileNotFoundException
    {
	putDir(dir, durl, null, true);
    }

    /**
     * Put (recursively) the files to the location URL.
     * @param dir the directory
     * @param durl the location URL
     * @param headers User headers
     */
    public void putDir(File dir, String durl, Hashtable headers, boolean rec) 
	throws FileNotFoundException
    {
	Vector dirs = new Vector(3);
	File files[] = dir.listFiles();
	for (int i = 0 ; i < files.length ; i++) {
	    if (rec && (files[i].isDirectory()))
		dirs.addElement(files[i]);
	    else if (! files[i].isDirectory()) {
		try {
		    putFile(files[i], durl+files[i].getName());
		} catch (Exception ex) {
		    ui.errlog(files[i].getName()+": "+ex.getMessage());
		}
	    }
	}
	ui.sync();
	for (int i = 0 ; i < dirs.size() ; i++) {
	    File subdir = (File)dirs.elementAt(i);
	    putDir(subdir, durl+subdir.getName()+"/", headers, rec);
	}
	String message[] = {
	    "Directory "+dir.getName()+" uploaded at",
	    durl
	};
	ui.displayMessage(message, "Directory uploaded");
    }

    
    /**
     * Put the givent file at the given location URL
     * @param toput the file to upload
     * @param furl the url
     */
    public void putFile(File toput, String furl) 
	throws FileNotFoundException
    {
	putFile(toput, furl, null);
    }

    /**
     * Put the givent file at the given location URL
     * @param toput the file to upload
     * @param furl the url
     * @param headers User headers
     */
    public void putFile(File toput, String desturl, Hashtable headers) 
	throws FileNotFoundException
    {
	String furl = null;
	if (desturl.endsWith("/")) {
	    furl = desturl+toput.getName();
	} else {
	    furl = desturl;
	}
	//always useful, at least to know if distant server is HTTP/1.1
	Reply reply = performRequest("HEAD", furl);
	if (reply == null)
	    return;
	skipBody(reply);
	String location = reply.getLocation();
	furl = (location != null ? location : furl);
	String etag = null;
	if (! overwrite) {
	    etag = getETag(toput, furl);
	    if (etag == null) { 
		//no etag, first time we put it
		int status = reply.getStatus();
		if (status/100 == 2) {
		    if (! ui.askQuestion(toput.getName()+
					 " exists on remote server.",
					 "Overwrite?"))
			return;
		}
	    } else {
		//we must compare etags
		HttpEntityTag remoteETag = reply.getETag();
		if ((remoteETag != null) && 
		    (! etag.equals(remoteETag.getTag()))) {
		    //conflict
		    String array1[] = { 
			toput.getName()+":",
			"Conflict!!! The remote file has been modified."};
		    String array2[] = {
			"Do you want to backup the local file",
			"and download the latest version?" };
		    boolean yes = ui.askQuestion(array1, array2);
		    if (yes) {
			try {
			    wget.getFile(toput, furl, true);
			} catch (IOException ex) {
			    ui.displayError(ex.getMessage());
			}
		    }
		    return;
		}
	    }
	}
	//here we can perform the PUT, and store the new etag.
	if (! quiet) {
	    ui.log("Uploading "+toput.getName()+"...");
	    ui.sync();
	}

	URL puturl  = null;
	try {
	    //check the url
	    puturl = new URL(furl);
	} catch (MalformedURLException ex) {
	    throw new RuntimeException("Invalid url : "+furl.toString());
	}
	//send the put request
	reply = put(toput, furl, etag, headers);

	if (reply == null)
	    return;
	if (! quiet) {
	    if (reply.getStatus() / 100 == 2) {
		String msg[] = { "File "+toput.getName()+" uploaded at",
				 furl };
		ui.displayMessage(msg, "File uploaded");
		//save etag
		HttpEntityTag remoteETag = reply.getETag();
		if (remoteETag != null)
		    saveEtag(toput, furl, remoteETag.getTag());
		skipBody(reply);
	    } else if (reply.getStatus() == HTTP.PRECONDITION_FAILED) {
		skipBody(reply);
		//the file has changed.
		boolean yes = 
		    ui.askQuestion("The file has been modified again!!!",
				   "Do you want to download the updated "+
				   "version?");
		if (yes) {
		    try {
			wget.getFile(toput, furl, false);
		    } catch (IOException ex) {
			ui.displayError(ex.getMessage());
		    }
		}
	    } else {
		//error
		displayErrorMessage(reply);
		if (!rec && furl.endsWith("/")) {
		    String message[] = 
		    { "WARNING : Your destination url ends with a '/'",
		      "The destination url must be the full url ",
		      "(with the filename)." };
		    ui.displayError(message, "Warning!");
		}
	    }
	} else {
	    skipBody(reply);
	}
    }

    /**
     * Constructor
     * @param ui the User Interface
     */
    public JWPut(JWUI ui) {
	this(null, ui, new Properties());
    }

    /**
     * Constructor
     * @param ui the User Interface
     * @param props the User properties
     */
    public JWPut(JWUI ui, Properties props) {
	this(null, ui, props);
    }

    /**
     * Constructor.
     * @param file The file (or directory) to put
     * @param mime The mime type
     * @param url The destination URL
     * @param ui The User Interface
     * @param userprops User properties
     */
    public JWPut(String mime, Properties userprops) 
    {
	this(mime, new org.w3c.jwput.tui.JWTUI(), userprops);
    }
    /**
     * Constructor.
     * @param file The file (or directory) to put
     * @param mime The mime type
     * @param url The destination URL
     * @param userprops User properties
     */
    public JWPut(String mime, JWUI ui, Properties userprops) 
    {
	this.mime = mime;
	this.ui   = ui;

	Properties props   = new Properties();
	File file = getPreferencesFile();
	try {
	    InputStream in = 
		new BufferedInputStream(new FileInputStream(file));
	    try {
		props.load(in);
	    } catch (IOException ex) {
		ui.displayError("unable to load "+WPUTRC);
		ui.errlog(ex.getMessage());
		ui.sync();
	    }
	} catch (FileNotFoundException ex) {
	    ui.errlog(file.getAbsolutePath()+" not found");
	}
	//merge user properties with default properties
	Enumeration enum = userprops.propertyNames();
	while (enum.hasMoreElements()) {
	    String key = (String)enum.nextElement();
	    props.put(key, userprops.getProperty(key));
	}
	initialize(props);
	this.manager = getManager(props);
	this.wget = new JWGet(this.manager, ui);
    }

    public void updateProperties(Properties props) {
	super.updateProperties(props);
	this.manager = getManager(props);
	this.wget    = new JWGet(this.manager, ui);
    }

    protected void initialize(Properties props) {
	super.initialize(props);

	String over = props.getProperty(OVERWRITE_P);
	this.overwrite = 
	    (over == null) ? false : over.equalsIgnoreCase(TRUE_P);

	String recs = props.getProperty(REC_P);
	this.rec = 
	    (recs == null) ? false : recs.equalsIgnoreCase(TRUE_P);

    }

    /**
     * Display the "Usage" message.
     */
    public static void usage() {
	System.out.println("JWPut "+VERSION+", a HTTP/1.1 PUT tool.");
	System.out.println("Usage: java JWPut [OPTION] [file] <url>");
	System.out.println("\nOptions:\n");
	System.out.println("  -delete                delete the remote file.");
	System.out.println("  -h                     display this message.");
	System.out.println("  -header_XXX <value>    specify a HTTP header");
	System.out.println("  -mime <mime-type>      Specify the mime type.");
	System.out.println("  -overwrite             always overwrite remote "+
			   "files.");
	System.out.println("  -no-proxy              direct connection.");
	System.out.println("  -prompt                prompt monde on.");
	System.out.println("  -user <username>       Auth: the user name.");
	System.out.println("  -password <password>   Auth: the password.");
	System.out.println("  -proxy-host <host>     eg: proxy.foobar.com");
	System.out.println("  -proxy-port <port>     eg: 81 (default value)");
	System.out.println("  -quiet                 quiet mode on.");
	System.out.println("  -r                     recursive put.");
	System.out.println("  -verbose               verbose mode on.");
	System.out.println("");
	System.out.println("\nMail bug reports to <bmahe@w3.org>.");
	System.exit(1);
    }

    /**
     * Display the message for bad command line error.
     * @param msg a message.
     */
    public static void badArgs(String msg) {
	System.out.println("JWPut: "+msg);
	System.out.println("Usage: java JWPut [OPTION] [file] <url>");
	System.out.println("\n'JWPut -h' for more options.");
	System.exit(1);
    }

    /**
     * Main. read the command line.
     */
    public static void main(String args[]) {
	String  mime = null;
	String  file = null;
	String  url  = null;
	boolean del  = false;

	Properties props = new Properties();
	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-mime")) {
		mime = args[++i];
	    } else if (args[i].equals("-user")) {
		props.put(USER_P, args[++i]);
	    } else if (args[i].equals("-password")) {
		props.put(PASSWORD_P, args[++i]);
	    } else if (args[i].equals("-no-proxy")) {
		props.put(PROXY_SET_P, FALSE_P);
	    } else if (args[i].equals("-proxy-host")) {
		props.put(PROXY_HOST_P, args[++i]);
	    } else if (args[i].equals("-proxy-port")) {
		props.put(PROXY_PORT_P, args[++i]);
	    } else if (args[i].equals("-verbose")) {
		props.put(VERBOSE_P, TRUE_P);
	    } else if (args[i].equals("-overwrite")) {
		props.put(OVERWRITE_P, TRUE_P);
	    } else if (args[i].equals("-quiet")) {
		props.put(QUIET_P, TRUE_P);
	    } else if (args[i].equals("-r")) {
		props.put(REC_P, TRUE_P);
	    } else if (args[i].equals("-prompt")) {
		props.put(PROMPT_P, TRUE_P);
	    } else if (args[i].equals("-delete")) {
		del = true;
	    } else if (args[i].equals("-h")) {
		usage();
	    } else if (args[i].startsWith("-"+HEADER_PREFIX)) {
		props.put(args[i].substring(1), args[++i]);
	    } else {
		if (del) {
		    if (url == null)
			url = args[i];
		    else
			usage();
		} else {
		    if (file == null)
			file = args[i];
		    else if (url == null)
			url = args[i];
		    else
			usage();
		}
	    }
	}
	if ((! del) && (file == null))
	    badArgs("Missing file");
	else if (url == null)
	    badArgs("Mising URL");


	JWPut wput = new JWPut(mime, props);
	try {
	    if (del)
		wput.delete(url);
	    else
		wput.perform(file, url);
	} catch (Exception ex) {
	    System.err.println(ex.getMessage());
	}
    }
}
